/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.ws.dbsources;

import jalview.datamodel.Alignment;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.DBRefEntry;
import jalview.datamodel.DBRefSource;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.Sequence;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.datamodel.UniprotEntry;
import jalview.datamodel.UniprotFile;
import jalview.ws.ebi.EBIFetchClient;
import jalview.ws.seqfetcher.DbSourceProxy;
import jalview.ws.seqfetcher.DbSourceProxyImpl;

import java.io.File;
import java.io.FileReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Vector;

import org.exolab.castor.xml.Unmarshaller;

import com.stevesoft.pat.Regex;

/**
 * @author JimP
 * 
 */
public class Uniprot extends DbSourceProxyImpl implements DbSourceProxy
{

  private static final String BAR_DELIMITER = "|";

  private static org.exolab.castor.mapping.Mapping map;

  /**
   * Constructor
   */
  public Uniprot()
  {
    super();
    addDbSourceProperty(DBRefSource.SEQDB, DBRefSource.SEQDB);
    addDbSourceProperty(DBRefSource.PROTSEQDB);
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#getAccessionSeparator()
   */
  @Override
  public String getAccessionSeparator()
  {
    return null;
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#getAccessionValidator()
   */
  @Override
  public Regex getAccessionValidator()
  {
    return new Regex("([A-Z]+[0-9]+[A-Z0-9]+|[A-Z0-9]+_[A-Z0-9]+)");
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#getDbSource()
   */
  @Override
  public String getDbSource()
  {
    return DBRefSource.UNIPROT;
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#getDbVersion()
   */
  @Override
  public String getDbVersion()
  {
    return "0"; // we really don't know what version we're on.
  }

  /**
   * Reads a file containing the reply to the EBI Fetch Uniprot data query,
   * unmarshals it to a UniprotFile object, and returns the list of UniprotEntry
   * data models (mapped from &lt;entry&gt; elements)
   * 
   * @param fileReader
   * @return
   */
  public Vector<UniprotEntry> getUniprotEntries(Reader fileReader)
  {
    UniprotFile uni = new UniprotFile();
    try
    {
      if (map == null)
      {
        // 1. Load the mapping information from the file
        map = new org.exolab.castor.mapping.Mapping(uni.getClass()
                .getClassLoader());
        java.net.URL url = getClass().getResource("/uniprot_mapping.xml");
        map.loadMapping(url);
      }

      // 2. Unmarshal the data
      Unmarshaller unmar = new Unmarshaller(uni);
      unmar.setIgnoreExtraElements(true);
      unmar.setMapping(map);
      if (fileReader != null)
      {
        uni = (UniprotFile) unmar.unmarshal(fileReader);
      }
    } catch (Exception e)
    {
      System.out.println("Error getUniprotEntries() " + e);
    }

    return uni.getUniprotEntries();
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#getSequenceRecords(java.lang.String[])
   */
  @Override
  public AlignmentI getSequenceRecords(String queries) throws Exception
  {
    startQuery();
    try
    {
      queries = queries.toUpperCase().replaceAll(
              "(UNIPROT\\|?|UNIPROT_|UNIREF\\d+_|UNIREF\\d+\\|?)", "");
      AlignmentI al = null;
      EBIFetchClient ebi = new EBIFetchClient();
      // uniprotxml parameter required since december 2007
      // uniprotkb dbname changed introduced december 2008
      File file = ebi.fetchDataAsFile("uniprotkb:" + queries, "uniprotxml",
              null);
      Vector<UniprotEntry> entries = getUniprotEntries(new FileReader(file));

      if (entries != null)
      {
        ArrayList<SequenceI> seqs = new ArrayList<SequenceI>();
        for (UniprotEntry entry : entries)
        {
          seqs.add(uniprotEntryToSequenceI(entry));
        }
        al = new Alignment(seqs.toArray(new SequenceI[0]));

      }
      stopQuery();
      return al;
    } catch (Exception e)
    {
      stopQuery();
      throw (e);
    }
  }

  /**
   * 
   * @param entry
   *          UniprotEntry
   * @return SequenceI instance created from the UniprotEntry instance
   */
  public SequenceI uniprotEntryToSequenceI(UniprotEntry entry){
    String id = getUniprotEntryId(entry);
    SequenceI sequence = new Sequence(id, entry.getUniprotSequence()
            .getContent());
    sequence.setDescription(getUniprotEntryDescription(entry));

    final String dbVersion = getDbVersion();
    ArrayList<DBRefEntry> dbRefs = new ArrayList<DBRefEntry>();
    for (String accessionId : entry.getAccession())
    {
      DBRefEntry dbRef = new DBRefEntry(DBRefSource.UNIPROT, dbVersion,
              accessionId);
      dbRefs.add(dbRef);
    }
    sequence.setSourceDBRef((dbRefs != null && dbRefs.size() > 0) ? dbRefs
            .get(0) : null);

    Vector<PDBEntry> onlyPdbEntries = new Vector<PDBEntry>();
    for (PDBEntry pdb : entry.getDbReference())
    {
      DBRefEntry dbr = new DBRefEntry();
      dbr.setSource(pdb.getType());
      dbr.setAccessionId(pdb.getId());
      dbr.setVersion(DBRefSource.UNIPROT + ":" + dbVersion);
      dbRefs.add(dbr);
      if ("PDB".equals(pdb.getType()))
      {
        onlyPdbEntries.addElement(pdb);
      }
    }

    sequence.setPDBId(onlyPdbEntries);
    if (entry.getFeature() != null)
    {
      for (SequenceFeature sf : entry.getFeature())
      {
        sf.setFeatureGroup("Uniprot");
        sequence.addSequenceFeature(sf);
      }
    }
    sequence.setDBRefs(dbRefs.toArray(new DBRefEntry[0]));
    return sequence;
  }

  /**
   * 
   * @param entry
   *          UniportEntry
   * @return protein name(s) delimited by a white space character
   */
  public static String getUniprotEntryDescription(UniprotEntry entry)
  {
    StringBuilder desc = new StringBuilder(32);
    if (entry.getProtein() != null && entry.getProtein().getName() != null)
    {
      for (String nm : entry.getProtein().getName())
      {
        desc.append(nm).append(" ");
      }
    }
    return desc.toString();
  }

  /**
   *
   * @param entry
   *          UniportEntry
   * @return The accession id(s) and name(s) delimited by '|'.
   */
  public static String getUniprotEntryId(UniprotEntry entry)
  {
    StringBuilder name = new StringBuilder(32);
    name.append("UniProt/Swiss-Prot");
    for (String accessionId : entry.getAccession())
    {
      name.append(BAR_DELIMITER);
      name.append(accessionId);
    }
    for (String n : entry.getName())
    {
      name.append(BAR_DELIMITER);
      name.append(n);
    }
    return name.toString();
  }

  /*
   * (non-Javadoc)
   * 
   * @see jalview.ws.DbSourceProxy#isValidReference(java.lang.String)
   */
  @Override
  public boolean isValidReference(String accession)
  {
    // TODO: make the following a standard validator
    return (accession == null || accession.length() < 2) ? false
            : getAccessionValidator().search(accession);
  }

  /**
   * return LDHA_CHICK uniprot entry
   */
  @Override
  public String getTestQuery()
  {
    return "P00340";
  }

  @Override
  public String getDbName()
  {
    return "Uniprot"; // getDbSource();
  }

  @Override
  public int getTier()
  {
    return 0;
  }
}
